/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cinterfacetype.cc
 *		- abstract class for interface types 
 *
 *
 ******************************************************************************
 */

#include <list>
#include "cinterfacetype.h"
#include "cvar.h"
#include "cnode.h"

/**
 * Create interface type instance
 */
CInterfaceType::CInterfaceType( CSymbol* symbol, CSymbol* modportSymbol, list<CNode*>& rangeList )
  : CDataType(eINTERFACE_TYPE, rangeList), 
    symbol(symbol),
    modportSymbol(modportSymbol)
{
}
/**
 * Create a new copy with a deep copy. 
 * \param heap heap to use for allocation.
 */
CDataType* CInterfaceType::Clone( CObstack* heap )
{
    list<CNode*> emptyList;
    CInterfaceType* clone = new(heap) CInterfaceType(GetSymbol(), GetModportSymbol(), emptyList);
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CInterfaceType::Copy( CObstack* heap, const CInterfaceType& o )
{
    CDataType::Copy( heap, o );
}

/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CInterfaceType::Dump( FILE* f ) const
{
    fprintf( f, "interface " );
}

/**
 * Get declaration's 2 state property.
 * \return value of 2 state property.
 */
static int _GetTwoState( CNode* n, void* data )
{
    int* twoState = static_cast<int*>(data);
    switch( n->GetOp() ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*twoState &= n->Arg<CDecl*>(0)->GetDataType()->GetTwoState();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CInterfaceType::GetTwoState() const
{
    return false;
}
/**
 * Determine if packed or vector width of declaration is constant, 
 * ie dependent upon only constants and parameters.
 * \return non-zero if vector or packed width of declaration 
 * is constant.
 */
static int _IsWidthConstant( CNode* n, void* data )
{
    int* isWidthConstant = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*isWidthConstant &= n->Arg<CDecl*>(0)->IsWidthConstant();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CInterfaceType::IsBaseWidthConstant() const
{
    return false;
}

/**
 * Determine if packed or vector width of declaration is volatile, 
 * ie depend upon parameters or variables.
 * \return non-zero of vector or packed width of declaration 
 * is volatile.
 */
static int _IsWidthVolatile( CNode* n, void* data )
{
    int* isWidthVolatile = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*isWidthVolatile |= n->Arg<CDecl*>(0)->IsWidthVolatile();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
int CInterfaceType::IsBaseWidthVolatile() const
{
    return false;
}
/**
 * Determine if packed or vector width of declaration can be evaluated.
 * \return non-zero of packed or vector width of declaration can 
 * be evaluated.
 */
static int _IsWidthEvaluateable( CNode* n, void* data )
{
    int* isWidthEvaluateable = static_cast<int*>(data);
    NodeOp_t t = n->GetOp();
    switch( t ) {
    case eLIST:
        break;
    case ePARAM_DECL:
    case eVAR_DECL:
        *isWidthEvaluateable &= n->Arg<CDecl*>(0)->IsWidthEvaluateable();
        return 0;
    default:
        MASSERT( FALSE );
    }
    return 1;
}
int CInterfaceType::IsBaseWidthEvaluateable() const
{
    return false;
}

/**
 * Evaluate packed or vector width of declaration.
 * \return packed or vector width of declaration.
 */
static int _GetWidthStruct( CNode* n, void* data )
{
    int* width = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*width += n->Arg<CDecl*>(0)->GetWidth();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
static int _GetWidthUnion( CNode* n, void* data )
{
    int* width = static_cast<int*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL:
	*width = (*width > n->Arg<CDecl*>(0)->GetWidth()) ? *width : 
			n->Arg<CDecl*>(0)->GetWidth();
	return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
INT32 CInterfaceType::GetBaseWidth( void ) const
{
    return 0;
}

/**
 * Get expression for datatype's overall packed or vector width
 * \return width expression
 */
static int _GetWidthExpStruct( CNode* n, void* data )
{
    CNode** width = static_cast<CNode**>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CNode* exp = n->Arg<CDecl*>(0)->GetWidthExp();
	if( *width ) {
	    *width = cADD(*width,exp);
	} else {
	    *width = exp;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
static int _GetWidthExpUnion( CNode* n, void* data )
{
    CNode** width = static_cast<CNode**>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CNode* exp = n->Arg<CDecl*>(0)->GetWidthExp();
	if( *width ) {
	    *width = cMAX(*width,exp);
	} else {
	    *width = exp;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
CNode* CInterfaceType::GetBaseWidthExp() const
{
    return NULL;
}
/**
 * Find declaration for member
 * \param member  member to search for.
 * \return matching declaration, NULL if not found
 */

struct FindMemberInfo {
    CDecl*   decl;
    CSymbol* member;
};
static int _FindMember( CNode* n, void* data ) 
{
    FindMemberInfo* info = static_cast<FindMemberInfo*>(data);
    switch( n->GetOp()  ) {
    case eLIST:
	break;
    case ePARAM_DECL:
    case eVAR_DECL: {
	CDecl* decl = n->Arg<CDecl*>(0);
	if( decl->GetSymbol() == info->member ) {
	    info->decl = decl;
	}
	} return 0;
    default:
	MASSERT( FALSE );
    }
    return 1;
}
CDecl* CInterfaceType::FindMember( CSymbol* member )
{
    MASSERT( false );
    return NULL; 
}
/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CInterfaceType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
}
/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CInterfaceType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
}

/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CInterfaceType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
}
